import 'dart:ffi';

import 'package:dio/dio.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:handy_ui_kit/handy_ui_kit.dart';

/// A Network Manager.
class NetworkManager {
  /// have several http functions

  ///called on start every request
  Function onStartDefault = () {};

  ///called on End every request
  Function onEndDefault = () {};

  ///called on Success every request
  Function onSuccessDefault = () {};

  ///called on Failed every request
  Function onFailedDefault = () {};

  ///need context to show dialog or pop and push
  BuildContext context;

  ///called on start this instance of Network Manager request
  Function onStart = () {};

  ///called on end this instance of Network Manager request
  Function onEnd = () {};

  ///called on success this instance of Network Manager request
  void Function(BuildContext context, dynamic responseJson) onSuccess = (BuildContext context, dynamic responseJson) {};

  ///called on failed this instance of Network Manager request
  void Function(BuildContext context, dynamic responseJson) onFailed = (BuildContext context, dynamic responseJson) {};

  ///other rules for check beside result code 200 should be a function return bool value
  bool Function(dynamic responseJson) successRules = (dynamic responseJson) => true;

  ///check to show dialog when when request is success
  bool showFailedDialog = false;

  ///check to show dialog when when request is failed
  bool showSuccessDialog = false;

  ///check to call default functions each step
  bool callDefaults = true;

  ///api address you're requesting
  String api;

  ///Map<String dynamic> request data for post methods
  Map<String, dynamic> req;

  ///Map<String dynamic> query data for get methods
  Map<String, dynamic> query;

  ///String token which is used in header
  String token = "";

  /// dio instance for making requests
  Dio dio = Dio()..options = BaseOptions(headers: {'Content-Type': 'application/json'});

  /// a function to extract error as string
  String Function(dynamic error) errorMsgExtractor = (error) => error.toString();
  String Function(dynamic msg) successMsgExtractor = (msg) => msg.toString();

  static final NetworkManager _singleton = NetworkManager._internal();

  factory NetworkManager(
      {@required BuildContext context,
      @required String api,
      Map<String, dynamic> req,
      Map<String, dynamic> query,
      void Function(BuildContext context, dynamic result) onSuccess,
      void Function(BuildContext context, dynamic result) onFailed,
      Function onStart,
      Function onEnd,
      bool showFailedDialog = false,
      bool callDefaults = true,
      bool showSuccessDialog = false}) {
    _singleton.context = context;
    _singleton.onStart = onStart ?? () {};
    _singleton.onEnd = onEnd ?? () {};
    _singleton.onSuccess = onSuccess ?? (BuildContext context, dynamic responseJson) {};
    _singleton.onFailed = onFailed ?? (BuildContext context, dynamic responseJson) {};
    _singleton.req = req;
    _singleton.api = api;
    _singleton.query = query;
    _singleton.callDefaults = callDefaults;
    return _singleton;
  }

  NetworkManager._internal();

  static setAccessToken(String token) {
    _singleton.token = token;
  }

  static initialize({
    String baseUrl,
    int timeout = 3000,
    Function onStartDefault,
    Function onEndDefault,
    Function onSuccessDefault,
    Function onFailedDefault,
    Map<String, dynamic> headers,
    bool Function(dynamic responseJson) successRules,
    String Function(dynamic error) errorMsgExtractor,
    String Function(dynamic msg) successMsgExtractor,
  }) {
    _singleton.onStartDefault = onStartDefault ?? () {};
    _singleton.onEndDefault = onEndDefault ?? () {};
    _singleton.onSuccessDefault = onSuccessDefault ?? () {};
    _singleton.onFailedDefault = onFailedDefault ?? () {};
    _singleton.successRules = successRules ?? (dynamic responseJson) => true;
    _singleton.dio.options = BaseOptions(

        /// If the `path` starts with "http(s)", the `baseURL` will be ignored, otherwise,
        baseUrl: baseUrl,
        headers: headers ?? _singleton.dio.options.headers,

        /// Timeout in milliseconds for opening  url.
        connectTimeout: timeout,
        sendTimeout: timeout,
        receiveTimeout: timeout);

    ///You can set LogInterceptor to print request/response log automaticlly
//    _singleton.dio.interceptors.add(LogInterceptor(responseBody: false));
    if (errorMsgExtractor != null) {
      _singleton.errorMsgExtractor = errorMsgExtractor;
    }
    if (successMsgExtractor != null) {
      _singleton.successMsgExtractor = errorMsgExtractor;
    }
  }

  _requestPost() {
    if (callDefaults) {
      onStartDefault();
    }
    onStart();
    print("Posting $req in $api");
    dio.post(api, data: req).then((response) {
      var statusCode = response.statusCode;
      var resultJson = response.data;
      print("***Status Code*** : $statusCode");
      print("***Response Data*** : $resultJson");
      if (statusCode == 200) {
        if (successRules(resultJson)) {
          if (callDefaults) {
            onSuccessDefault();
          }
          onSuccess(context, resultJson);
          if (showSuccessDialog) {
            showDialog(
                context: context,
                builder: (context) => MyDialog(
                      title: "Success",
                      body: successMsgExtractor(resultJson),
                      type: DialogType.success,
                    ));
          }
        } else {
          if (callDefaults) {
            onFailedDefault();
          }
          onFailed(context, errorMsgExtractor(resultJson));
          if (showFailedDialog) {
            showDialog(
                context: context,
                builder: (context) => MyDialog(
                      title: "Failed",
                      body: errorMsgExtractor(resultJson),
                      type: DialogType.warning,
                    ));
          }
        }
      } else {
        if (callDefaults) {
          onFailedDefault();
        }
        onFailed(context, resultJson);
        if (showFailedDialog) {
          showDialog(
              context: context,
              builder: (context) => MyDialog(
                    title: "Failed",
                    body: resultJson.toString(),
                    type: DialogType.error,
                  ));
        }
      }
      if (callDefaults) {
        onEndDefault();
      }
      onEnd();
    }).catchError((e) {
      print("Error: $e");
      if (callDefaults) {
        onFailedDefault();
        onEndDefault();
      }
      onFailed(context, e);
      onEnd();
      if (showFailedDialog) {
        showDialog(
            context: context,
            builder: (context) => MyDialog(
                  title: "Failed",
                  body: e.toString(),
                  type: DialogType.error,
                ));
      }
    });
  }

  _requestGet() {
    if (callDefaults) {
      onStartDefault();
    }
    onStart();
    print("Getting $api with $query");
    dio.get(api, queryParameters: req).then((response) {
      var statusCode = response.statusCode;
      var resultJson = response.data;
      print("***Status Code*** : $statusCode");
      print("***Response Data*** : $resultJson");
      if (statusCode == 200 && successRules(resultJson)) {
        if (callDefaults) {
          onSuccessDefault();
        }
        onSuccess(context, resultJson);
      } else {
        if (callDefaults) {
          onFailedDefault();
        }
        onFailed(context, resultJson);
      }
      if (callDefaults) {
        onEndDefault();
      }
      onEnd();
    }).catchError((e) {
      if (callDefaults) {
        onFailedDefault();
        onEndDefault();
      }
      print("Error: $e");
      onFailed(context, e);
      onEnd();
    });
  }

  get requestPost => _requestPost;

  get requestGet => _requestGet;
}
